# Copyright (C) 2012 Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import time
import logging
import urllib2
from ConfigParser import ConfigParser
from tempfile import NamedTemporaryFile
from os.path import exists, join

from rpm import TransactionSet
from rpm import RPMCALLBACK_INST_OPEN_FILE, RPMCALLBACK_INST_CLOSE_FILE
from rpm import _RPMVSF_NODIGESTS, _RPMVSF_NOSIGNATURES

from sugar_network.client import IPCClient, solver
from sugar_network.toolkit.util import parse_version


_REPO_LIST = '/etc/PackageKit/presolve-repos'
_CACHE_DIR = '/var/cache/presolve'
_CACHE_EXPIRE = 60 * 5
_BUFFER_SIZE = 1024 * 100

_conn = IPCClient()
_ts = TransactionSet()


def resolve(package):
    """
    >>> _ts.dbMatch = lambda *args: [{'version': 1, 'arch': 'x86'}]
    >>> json.dumps(resolve('vte'))
    '{"version": 1, "arch": "x86", "installed": true}'

    >>> _ts.dbMatch = lambda *args: []
    >>> json.dumps(resolve('vte'))
    '{"version": "0.26.1", "arch": "i686", "installed": false}'

    """
    for rpm in _ts.dbMatch('name', package):
        logging.info('Resolved installed %r', package)
        return {'version': rpm['version'],
                'arch': rpm['arch'],
                'installed': True,
                }

    try:
        arch, binaries = _resolve(package)
    except Exception:
        logging.exception('Cannot resolve %r', package)
        return None

    logging.info('Resolved available %r: %r', package, binaries)

    return {'version': binaries[0]['version'],
            'arch': arch,
            'installed': False,
            }


def install(package):
    """
    >>> installs = []
    >>> _ts.run = lambda _, rpms: installs.append(len(rpms))

    >>> _ts.dbMatch = lambda *args: [{'name': 'vte', 'version': '1', 'release': '0'}]
    >>> install('vte')
    >>> installs
    []

    >>> _ts.dbMatch = lambda *args: [{'name': 'vte', 'version': '0.1', 'release': '0'}]
    >>> install('vte')
    >>> installs
    [1]

    """
    status = _conn.get(cmd='status')
    if status['route'] == 'offline':
        raise RuntimeError('No server connection')
    __, binaries = _resolve(package)

    rpms = []
    files = set()
    for newrpm in reversed(binaries):
        oldrpm = None
        for oldrpm in _ts.dbMatch('name', newrpm['name']):
            break
        if oldrpm is not None and _cmp_versions(newrpm, oldrpm) <= 0:
            continue

        url = newrpm.get('url')
        if not url:
            url = status['node'] + newrpm['path']
        filename = url.rsplit('/', 1)[-1]

        if filename in files:
            continue
        files.add(filename)

        logging.info('Download %r package', url)
        rpm = NamedTemporaryFile()
        dl = urllib2.urlopen(url)
        while True:
            chunk = dl.read(_BUFFER_SIZE)
            if not chunk:
                break
            rpm.write(chunk)
        rpm.flush()
        rpm.seek(0)
        rpms.append(rpm)

    if not rpms:
        return

    # Keys might be missed
    _ts.setVSFlags(_RPMVSF_NODIGESTS | _RPMVSF_NOSIGNATURES)
    for i, rpm in enumerate(rpms):
        header = _ts.hdrFromFdno(rpm.file.fileno())
        _ts.addInstall(header, i, 'i')
        rpm.seek(0)
    logging.info('Start installation')
    _ts.run(_install_callback, rpms)


def _cmp_versions(r1, r2):
    """
    >>> _cmp_versions({'version': '0.1', 'release': '1.fc14'}, {'version': '0.1', 'release': '1.fc14'})
    0

    >>> _cmp_versions({'version': '0.1', 'release': '1.fc14'}, {'version': '0.1', 'release': '2.fc14'})
    -1
    >>> _cmp_versions({'version': '0.1', 'release': '2.fc14'}, {'version': '0.1', 'release': '1.fc14'})
    1

    >>> _cmp_versions({'version': '0.1', 'release': '1.fc14'}, {'version': '0.2', 'release': '1.fc14'})
    -1
    >>> _cmp_versions({'version': '0.2', 'release': '1.fc14'}, {'version': '0.1', 'release': '1.fc14'})
    1

    >>> _cmp_versions({'version': '0.1', 'release': '1.fc14'}, {'epoch': '1', 'version': '0.1', 'release': '1.fc14'})
    -1
    >>> _cmp_versions({'epoch': '1', 'version': '0.1', 'release': '1.fc14'}, {'version': '0.1', 'release': '1.fc14'})
    1

    """
    d = cmp(parse_version(r1.get('epoch') or '0'),
            parse_version(r2.get('epoch') or '0'))
    if d:
        return d
    d = cmp(parse_version(r1['version']), parse_version(r2['version']))
    if d:
        return d

    def parse_release(r):
        return parse_version(r['release'].rsplit('.', 1)[0])

    return cmp(parse_release(r1), parse_release(r2))


def _resolve(package):
    cache_path = join(_CACHE_DIR, package)
    if exists(cache_path) and \
            (time.time() - os.stat(cache_path).st_mtime) < _CACHE_EXPIRE:
        with file(cache_path) as f:
            return json.load(f)

    if not exists(_CACHE_DIR):
        os.makedirs(_CACHE_DIR)

    for repo in file(_REPO_LIST).read().split():
        try:
            response = _conn.request('GET', ['packages', repo, package])
        except Exception:
            logging.exception('No %s in %s, try next repo', package, repo)
        else:
            break
    else:
        raise RuntimeError('Package not found')

    binaries = json.load(response.raw)

    arch = solver.select_architecture(binaries.keys())
    if not arch:
        raise RuntimeError('No binaries for the host architecture')
    binaries = binaries[arch]

    for binfile in binaries:
        path = binfile.get('path')
        if path and not path.startswith('/'):
            binfile['path'] = '/'.join(['', 'packages', repo, path])

    with file(cache_path, 'w') as f:
        json.dump([arch, binaries], f)

    return arch, binaries


def _install_callback(reason, amount, total, key, rpms):
    if reason == RPMCALLBACK_INST_OPEN_FILE:
        return rpms[key].file.fileno()
    elif reason == RPMCALLBACK_INST_CLOSE_FILE:
        rpms[key].close()


if __name__ == "__main__":
    import doctest
    import shutil

    root = '/tmp/%s/' % __file__
    shutil.rmtree(root, ignore_errors=True)
    os.makedirs(root)

    _CACHE_DIR = root
    _REPO_LIST = root + 'repos'
    with file(_REPO_LIST, 'w') as f:
        f.write('OLPC/11.3.1')

    logging.basicConfig(level=logging.DEBUG)
    #doctest.run_docstring_examples(resolve, globals())
    doctest.testmod()
